/**
 * @copyright 2017 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/*jshint browser: true *//* eslint-env browser */

/**
 * WebSocket Connection.
 *
 * @module env/WebSocketConnection
 * @private
 */
define([ "bajaScript/comm",
        "bajaScript/baja/comm/Callback",
        "bajaScript/env/Connection" ], function (
        baja, 
        Callback,
        Connection) {

  "use strict";

  /**
   * A connection that uses a WebSocket as a connection.
   *
   * @class
   * @private
   * @extends module:env/Connection
   * @alias module:env/WebSocketConnection
   * @param {Object} handlers Holds a collection of functions used as asynchronous callbacks.
   */
  var WebSocketConnection = function (handlers) {
    baja.callSuper(WebSocketConnection, this, arguments);
  };

  baja.subclass(WebSocketConnection, Connection);

  /**
   * Make a new server session. This must be done using the default HTTP
   * mechanism before a web socket can be opened.
   * @return {Promise} promise to be resolved after the server session is
   * created
   */
  WebSocketConnection.prototype.doAuthenticate = function () {
    var cb = new Callback();
    baja.comm.makeServerSession(cb);
    return cb.promise();
  };

  /**
   * Opens a new web socket to the station.
   */
  WebSocketConnection.prototype.start = function () {
    var that = this,
        handlers = that.$handlers,
        connection = that.$internalConnection = that.$createSocket();

    for (var name in handlers) {
      if (handlers.hasOwnProperty(name)) {
        connection.addEventListener(name, handlers[name], false);
      }
    }
  };

  /**
   * @private
   * @returns {WebSocket} web socket pointed at the BOX WebSocket acceptor
   */
  WebSocketConnection.prototype.$createSocket = function () {
    var isSecure = /^https/i.test(window.location.protocol),
      protocol = isSecure ? 'wss' : 'ws',
      uri = protocol + '://' + window.location.host + '/wsbox';

    // TODO: We should list supported protocols in the WebSocket constructor?
    // https://developer.mozilla.org/en-US/docs/WebSockets/Writing_WebSocket_client_applications

    try {
      return new WebSocket(uri);
    } catch (e) {
      baja.error('WebSockets are available in this browser, but could not ' +
        'construct a WebSocket connection to ' + uri);
      throw e;
    }
  };

  return WebSocketConnection;
});
